(******************************************************************************)
(* plx parser                                                                 *)
(* (C) 2010 Steffen Solyga <solyga@absinth.net>                               *)
(******************************************************************************)
IMPLEMENTATION MODULE Prsr;

IMPORT FIO;
FROM Scnr IMPORT Put, Read, KeepId, Diff, TSymbol, sym, num, id, pos;
FROM Gen IMPORT TValue, TAddress, TOp0, TOp1, TOp2, TItemDesc, TType;
FROM Gen IMPORT Hlt, Val, Sto, Lod, Op0, Op1, Op2;
FROM Gen IMPORT Label, Jmp, Jpf, Init, JmpFwd, JpfFwd, Fixup;
FROM Gen IMPORT vFalse, vTrue;

CONST
  err ::= FIO.ErrorOutput;
  cntMax = 32;

TYPE
  TClass = ( cConst, cVar, cType );
  TObjDesc = RECORD
    id : CARDINAL;
    typ: TType;
    CASE cls: TClass OF
    | cConst: val: TValue;
    | cVar  : idx: CARDINAL;
    | cType :
    END;
  END;

VAR
  objs: ARRAY [ 0 .. cntMax - 1 ] OF TObjDesc;
  cnt: CARDINAL;
  idx: CARDINAL;
  op2s: ARRAY TSymbol OF TOp2;


PROCEDURE Err( s: ARRAY OF CHAR );
BEGIN
  FIO.WrStr( err, 'Prsr (' );
  FIO.WrCard( err, pos.l, 0 );
  FIO.WrChar( err, ',' );
  FIO.WrCard( err, pos.c, 0 );
  FIO.WrStr( err, '): ' );
  FIO.WrStr( err, s );
  FIO.WrStr( err, '.' );
  FIO.WrLn( err );
  HALT;
END Err;

PROCEDURE Find( VAR i: CARDINAL );
BEGIN
  i := 0;
  WHILE i < cnt DO
    IF Diff( objs[i].id, id ) = 0 THEN RETURN END;
    INC( i );
  END;
END Find;

PROCEDURE ChkBool( VAR x: TItemDesc );
BEGIN
  IF x.t # tBool THEN Err( 'Not boolean' ) END;
END ChkBool;

PROCEDURE ChkInt( VAR x: TItemDesc );
BEGIN
  IF x.t # tInt THEN Err( 'Not integer' ) END;
END ChkInt;

PROCEDURE ChkCompat( VAR x, y: TItemDesc );
BEGIN
  IF x.t # y.t THEN Err( 'Incompatible types' ) END;
END ChkCompat;


PROCEDURE Expression( VAR x: TItemDesc );

  PROCEDURE SimpleExpr( VAR x: TItemDesc );

    PROCEDURE Factor( VAR x: TItemDesc );
    VAR i: CARDINAL;
    BEGIN
      CASE sym OF
      | sNum : IF num > MAX( TValue ) THEN Err( 'Number too large' ) END;
               x.t := tInt; x.c := TRUE; x.v := num;
               Read();
      | sId  : Find( i );
               IF i >= cnt THEN Err( 'Identifier undeclared' ) END;
               x.t := objs[ i ].typ;
               CASE objs[ i ].cls OF
               | cConst: x.c := TRUE; x.v := objs[ i ].val;
               | cVar  : x.c := FALSE; Lod( objs[i].idx );
               ELSE
                 Err( 'Expected constant or variable' );
               END;
               Read();
      | sLPar: Read(); Expression( x );
               IF sym # sRPar THEN Err( 'Expected operator or ")"' ) END;
               Read();
      | sNot : Read(); Factor( x ); ChkBool( x ); Op1( oNot, x );
      ELSE
        Err( 'Expected factor' );
      END;
    END Factor;

    PROCEDURE Term( VAR x: TItemDesc );
    VAR s: TSymbol; y: TItemDesc;
    BEGIN
      Factor( x );
      WHILE ( sMul <= sym ) & ( sym <= sAnd ) DO
        IF sym = sAnd THEN ChkBool( x ) ELSE ChkInt( x ) END;
        s := sym; Read(); Factor( y ); ChkCompat( x, y );
        Op2( op2s[s], x, y );
      END;
    END Term;

  VAR s: TSymbol; y: TItemDesc;
  BEGIN
    IF ( sym = sPlus ) OR ( sym = sMinus ) THEN
      s := sym; Read(); Term( x ); ChkInt( x );
      IF s = sMinus THEN Op1( oNeg, x ) END;
    ELSE
      Term( x );
    END;
    WHILE ( sPlus <= sym ) & ( sym <= sOr ) DO
      IF sym = sOr THEN ChkBool( x ) ELSE ChkInt( x ) END;
      s := sym; Read(); Term( y ); ChkCompat( x, y );
      Op2( op2s[s], x, y );
    END;
  END SimpleExpr;

VAR s: TSymbol; y: TItemDesc;
BEGIN
  SimpleExpr( x );
  IF ( sEql <= sym ) & ( sym <= sGeq ) THEN
    s := sym; Read(); SimpleExpr( y ); ChkCompat( x, y );
    Op2( op2s[s], x, y );
    x.t := tBool;
  END;
END Expression;

PROCEDURE ConstDeclaration();
VAR i: CARDINAL; x: TItemDesc;
BEGIN  (* sym = sId *)
  Find( i );
  IF i < cnt THEN Err( 'Identifier redeclared' ) END;
  IF cnt >= cntMax THEN Err( 'Too many identifiers' ) END;
  objs[ i ].id  := id;
  objs[ i ].cls := cConst;
  KeepId();
  Read();
  IF sym # sEql THEN Err( 'Expected "="' ) END;
  Read(); Expression( x );
  IF ~ x.c THEN Err( 'Expected constant' ) END;
  objs[ i ].typ := x.t;
  objs[ i ].val := x.v;
  INC( cnt );
  IF sym # sSemi THEN Err( 'Expected ";"' ) END;
  Read();
END ConstDeclaration;

PROCEDURE VarDeclaration();
VAR i,j: CARDINAL;
BEGIN  (* sym = sId *)
  Find( i );
  IF i < cnt THEN Err( 'Identifier redeclared' ) END;
  IF cnt >= cntMax THEN Err( 'Too many identifiers' ) END;
  objs[ i ].id  := id;
  objs[ i ].cls := cVar;
  KeepId();
  Read();
  IF sym # sCol THEN Err( 'Expected ":"' ) END;
  Read();
  IF sym # sId THEN Err( 'Expected identifier' ) END;
  Find( j );
  IF j >= cnt THEN Err( 'Identifier undeclared' ) END;
  IF objs[ j ].cls # cType THEN Err( 'Not a type' ) END;
  objs[ i ].typ := objs[ j ].typ;
  objs[ i ].idx := idx;
  INC( cnt ); INC( idx );
  Read();
  IF sym # sSemi THEN Err( 'Expected ";"' ) END;
  Read();
END VarDeclaration;

PROCEDURE StatementSequence();

  PROCEDURE Statement();
  VAR i: CARDINAL; x: TItemDesc; l, l1: TAddress;
  BEGIN
    CASE sym OF
    | sId: (* assignment *)
        Find( i );
        IF i >= cnt THEN Err( 'Identifier undeclared' ) END;
        IF objs[ i ].cls # cVar THEN Err( 'Not a variable' ) END;
        Read();
        IF sym # sBec THEN Err( 'Expected ":="' ) END;
        Read(); Expression( x );
        IF x.t # objs[i].typ THEN Err( 'Incompatible types' ) END;
        IF x.c THEN Val( x.v ) END;
        Sto( objs[i].idx );
    | sWrt: (* expression output *)
        Read(); Expression( x );
        IF x.c THEN Val( x.v ) END;
        CASE x.t OF
        | tInt : Op0( oWrt );
        | tBool: Op0( oWrB );
        ELSE
          Err( 'Bug: Invalid type' );
        END;
    | sRepeat:
        Read(); Label( l ); StatementSequence();
        IF sym # sUntil THEN Err( 'Expected "UNTIL"' ) END;
        Read(); Expression( x ); ChkBool( x ); Jpf( l, x );
    | sWhile:
        Read(); Label( l ); Expression( x ); ChkBool( x );
        IF sym # sDo THEN Err( 'Expected "DO"' ) END;
        Read(); Init( l1 ); JpfFwd( l1, x );
        StatementSequence(); Jmp( l ); Fixup( l1 );
        IF sym # sEnd THEN Err( 'Expected "END"' ) END;
        Read();
    | sIf:
        Read(); Expression( x ); ChkBool( x );
        Init( l ); JpfFwd( l, x );
        IF sym # sThen THEN Err( 'Expected "THEN"' ) END;
        Read(); StatementSequence();
        Init( l1 );
        WHILE sym = sElsif DO
          Read(); JmpFwd( l1 ); Fixup( l ); Expression( x ); ChkBool( x );
          Init( l ); JpfFwd( l, x );
          IF sym # sThen THEN Err( 'Expected "THEN"' ) END;
          Read(); StatementSequence();
        END;
        IF sym = sElse THEN
          Read(); JmpFwd( l1 ); Fixup( l ); StatementSequence();
        ELSE
          Fixup( l );
        END;
        Fixup( l1 );
        IF sym # sEnd THEN Err( '"IF" requires "END"' ) END;
        Read();
    ELSE
      (* empty *)
    END;
  END Statement;

BEGIN
  Statement();
  WHILE sym = sSemi DO Read(); Statement() END;
END StatementSequence;

PROCEDURE Run(); (* Program() *)
BEGIN
  Read();
  IF sym = sConst THEN
    Read();
    WHILE sym = sId DO ConstDeclaration() END;
  END;
  IF sym = sVar THEN
    Read();
    WHILE sym = sId DO VarDeclaration() END;
  END;
  IF sym # sBeg THEN Err( 'Expected "BEGIN"' ) END;
  Read(); StatementSequence();
  IF sym # sEnd THEN Err( '"BEGIN" requires "END"' ) END;
  Read();
  IF sym # sDot THEN Err( 'Expected ";" or "."' ) END;
(*  Read(); IF sym # sEOF THEN Err( 'Expected EOF' ) END; *)
  Hlt();
END Run;

PROCEDURE InitOprs();
VAR s: TSymbol;
BEGIN
  FOR s := MIN( TSymbol ) TO MAX( TSymbol ) DO op2s[ s ] := oNull END;
  op2s[ sMul   ] := oMul;
  op2s[ sDiv   ] := oDiv;
  op2s[ sAnd   ] := oAnd;
  op2s[ sPlus  ] := oAdd;
  op2s[ sMinus ] := oSub;
  op2s[ sOr    ] := oOr;
  op2s[ sEql   ] := oEql;
  op2s[ sNeq   ] := oNeq;
  op2s[ sLss   ] := oLss;
  op2s[ sLeq   ] := oLeq;
  op2s[ sGtr   ] := oGtr;
  op2s[ sGeq   ] := oGeq;
END InitOprs;

BEGIN
  cnt := 0;
  idx := 0;
  Put( 'FALSE' );
  objs[ cnt ].id  := id;
  objs[ cnt ].typ := tBool;
  objs[ cnt ].cls := cConst;
  objs[ cnt ].val := vFalse;
  KeepId(); INC( cnt );
  Put( 'TRUE' );
  objs[ cnt ].id  := id;
  objs[ cnt ].typ := tBool;
  objs[ cnt ].cls := cConst;
  objs[ cnt ].val := vTrue;
  KeepId(); INC( cnt );
  Put( 'BOOLEAN' );
  objs[ cnt ].id  := id;
  objs[ cnt ].typ := tBool;
  objs[ cnt ].cls := cType;
  KeepId(); INC( cnt );
  Put( 'INTEGER' );
  objs[ cnt ].id  := id;
  objs[ cnt ].typ := tInt;
  objs[ cnt ].cls := cType;
  KeepId(); INC( cnt );
  InitOprs();
END Prsr.
